"""Unit tests for ConditionTraceEngine."""

from __future__ import annotations

from typing import Any

from pyrung.core import Block, Bool, Int, TagType, all_of, any_of
from pyrung.core.condition import (
    BitCondition,
    Condition,
    FallingEdgeCondition,
    IntTruthyCondition,
    NormallyClosedCondition,
    RisingEdgeCondition,
)
from pyrung.core.condition_trace import ConditionTraceEngine
from pyrung.core.context import ConditionView, ScanContext
from pyrung.core.state import SystemState
from pyrung.core.trace_formatter import TraceFormatter


def _ctx(
    tags: dict[str, bool | int | float | str] | None = None,
    memory: dict[str, Any] | None = None,
) -> ScanContext:
    state = SystemState()
    if tags:
        state = state.with_tags(tags)
    if memory:
        state = state.with_memory(memory)
    return ScanContext(state)


def _detail_map(details: list[dict[str, Any]]) -> dict[str, Any]:
    return {str(item["name"]): item["value"] for item in details}


def test_condition_trace_engine_evaluates_basic_contacts_and_edges() -> None:
    engine = ConditionTraceEngine(formatter=TraceFormatter())
    button = Bool("Button")
    step = Int("Step")

    bit_cond = BitCondition(button)
    bit_value, bit_details = engine.evaluate(bit_cond, _ctx({"Button": True}))
    assert bit_value is True
    assert _detail_map(bit_details) == {"tag": "Button", "value": True}
    assert engine.expression(bit_cond) == "Button"

    int_cond = IntTruthyCondition(step)
    int_value, int_details = engine.evaluate(int_cond, _ctx({"Step": 2}))
    assert int_value is True
    assert _detail_map(int_details) == {"tag": "Step", "value": 2}
    assert engine.expression(int_cond) == "Step != 0"

    nc_cond = NormallyClosedCondition(button)
    nc_value, nc_details = engine.evaluate(nc_cond, _ctx({"Button": False}))
    assert nc_value is True
    assert _detail_map(nc_details) == {"tag": "Button", "value": False}
    assert engine.expression(nc_cond) == "!Button"

    rise_cond = RisingEdgeCondition(button)
    rise_value, rise_details = engine.evaluate(
        rise_cond,
        _ctx({"Button": True}, {"_prev:Button": False}),
    )
    assert rise_value is True
    assert _detail_map(rise_details) == {"tag": "Button", "current": True, "previous": False}
    assert engine.expression(rise_cond) == "rise(Button)"

    fall_cond = FallingEdgeCondition(button)
    fall_value, fall_details = engine.evaluate(
        fall_cond,
        _ctx({"Button": False}, {"_prev:Button": True}),
    )
    assert fall_value is True
    assert _detail_map(fall_details) == {"tag": "Button", "current": False, "previous": True}
    assert engine.expression(fall_cond) == "fall(Button)"


def test_condition_trace_engine_evaluates_direct_indirect_and_expr_comparisons() -> None:
    engine = ConditionTraceEngine(formatter=TraceFormatter())
    step = Int("Step")
    debug_step = Int("DebugStep")
    cur_step = Int("CurStep")
    step_block = Block(
        "Step",
        TagType.INT,
        0,
        9,
        address_formatter=lambda name, addr: f"{name}[{addr}]",
    )

    direct_cond = step == 1
    direct_value, direct_details = engine.evaluate(direct_cond, _ctx({"Step": 0}))
    assert direct_value is False
    assert _detail_map(direct_details) == {"left": "Step", "left_value": 0, "right_value": 1}
    assert engine.expression(direct_cond) == "Step == 1"

    left_indirect_cond = step_block[cur_step] == debug_step
    left_indirect_value, left_indirect_details = engine.evaluate(
        left_indirect_cond,
        _ctx({"CurStep": 1, "Step[1]": 0, "DebugStep": 5}),
    )
    assert left_indirect_value is False
    assert _detail_map(left_indirect_details) == {
        "left": "Step[1]",
        "left_value": 0,
        "right": "DebugStep",
        "right_value": 5,
        "left_pointer_expr": "Step[CurStep]",
        "left_pointer": "CurStep",
        "left_pointer_value": 1,
    }
    assert engine.expression(left_indirect_cond) == "Step[CurStep] == DebugStep"

    right_indirect_cond = debug_step == step_block[cur_step]
    right_indirect_value, right_indirect_details = engine.evaluate(
        right_indirect_cond,
        _ctx({"CurStep": 1, "Step[1]": 0, "DebugStep": 0}),
    )
    assert right_indirect_value is True
    assert _detail_map(right_indirect_details) == {
        "left": "DebugStep",
        "left_value": 0,
        "right": "Step[1]",
        "right_value": 0,
        "right_pointer_expr": "Step[CurStep]",
        "right_pointer": "CurStep",
        "right_pointer_value": 1,
    }
    assert engine.expression(right_indirect_cond) == "DebugStep == Step[CurStep]"

    expr_cond = (step + 1) > 0
    expr_value, expr_details = engine.evaluate(expr_cond, _ctx({"Step": 0}))
    assert expr_value is True
    expr_detail_map = _detail_map(expr_details)
    assert expr_detail_map["left"] == repr(expr_cond.left)
    assert expr_detail_map["left_value"] == 1
    assert expr_detail_map["right"] == repr(expr_cond.right)
    assert expr_detail_map["right_value"] == 0
    assert engine.expression(expr_cond) == f"{expr_cond.left!r} > {expr_cond.right!r}"


def test_condition_trace_engine_short_circuit_terms_and_composite_expressions() -> None:
    engine = ConditionTraceEngine(formatter=TraceFormatter())
    step = Int("Step")
    auto_mode = Bool("AutoMode")

    all_cond = all_of(step == 1, auto_mode)
    all_value, all_details = engine.evaluate(all_cond, _ctx({"Step": 0, "AutoMode": True}))
    assert all_value is False
    all_terms = str(_detail_map(all_details)["terms"])
    assert "Step(0) == 1(false)" in all_terms
    assert "AutoMode(skipped)" in all_terms
    assert engine.expression(all_cond) == "(Step == 1 & AutoMode)"

    any_cond = any_of(step == 0, auto_mode)
    any_value, any_details = engine.evaluate(any_cond, _ctx({"Step": 0, "AutoMode": False}))
    assert any_value is True
    any_terms = str(_detail_map(any_details)["terms"])
    assert "Step(0) == 0(true)" in any_terms
    assert "AutoMode(skipped)" in any_terms
    assert engine.expression(any_cond) == "(Step == 0 | AutoMode)"


def test_condition_trace_engine_unknown_condition_uses_fallbacks() -> None:
    class AlwaysTrueCondition(Condition):
        def evaluate(self, ctx: ScanContext | ConditionView) -> bool:
            _ = ctx
            return True

    engine = ConditionTraceEngine(formatter=TraceFormatter())
    cond = AlwaysTrueCondition()
    value, details = engine.evaluate(cond, _ctx())
    assert value is True
    assert details == []
    assert engine.expression(cond) == "AlwaysTrueCondition"
    assert engine.summary(cond, details) == "AlwaysTrueCondition"
    assert (
        engine.annotation(
            status="true",
            expression="AlwaysTrueCondition",
            summary="AlwaysTrueCondition",
        )
        == "[T] AlwaysTrueCondition"
    )



from __future__ import annotations

import inspect
import textwrap
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, ClassVar

from pyrung.core._source import (
    _capture_source,
    _capture_with_call_arg_lines,
    _capture_with_end_line,
)
from pyrung.core.condition import (
    ConditionTerm,
)
from pyrung.core.instruction import (
    ForLoopInstruction,
    SubroutineReturnSignal,
)
from pyrung.core.rung import Rung as RungLogic
from pyrung.core.tag import Tag, TagType

from .validation import _check_function_body_strict, _check_with_body_from_frame

if TYPE_CHECKING:
    from pyrung.core.context import ScanContext
    from pyrung.core.state import SystemState

    from .validation import DialectValidator

_rung_stack: list[Rung] = []


_forloop_active = False


def _current_rung() -> Rung | None:
    """Get the current rung context (if any)."""
    return _rung_stack[-1] if _rung_stack else None


def _require_rung_context(func_name: str) -> Rung:
    """Get current rung or raise error."""
    rung = _current_rung()
    if rung is None:
        raise RuntimeError(f"{func_name}() must be called inside a Rung context")
    return rung


def _push_rung_context(ctx: Rung) -> None:
    """Push a rung context onto the active stack."""
    _rung_stack.append(ctx)


def _pop_rung_context() -> None:
    """Pop the current rung context from the active stack."""
    _rung_stack.pop()


def _new_capture_context(rung: RungLogic) -> Rung:
    """Create a lightweight Rung wrapper for temporary capture scopes."""
    ctx = Rung.__new__(Rung)
    ctx._rung = rung
    ctx._pending_required_builder = None
    return ctx


def _set_scope_end_line(
    rung: RungLogic,
    *,
    source_file: str | None,
    source_line: int | None,
    context_name: str,
) -> None:
    """Populate end-line metadata for a scope-backed rung."""
    rung.end_line = _capture_with_end_line(
        source_file,
        source_line,
        context_name=context_name,
    )


def _validate_subroutine_name(name: str) -> str:
    if '"' in name:
        raise ValueError('Subroutine name must not contain ".')
    return name


class Program:
    """Container for PLC logic (rungs and subroutines).

    Used as a context manager to capture rungs:
        with Program() as logic:
            with Rung(Button):
                out(Light)

    Also works with PLCRunner:
        runner = PLCRunner(logic)
    """

    _current: Program | None = None
    _dialect_validators: ClassVar[dict[str, DialectValidator]] = {}

    def __init__(self, *, strict: bool = True) -> None:
        self._strict = strict
        self.rungs: list[RungLogic] = []
        self.subroutines: dict[str, list[RungLogic]] = {}
        self._current_subroutine: str | None = None  # Track if we're in a subroutine
        self._pending_comment: str | None = None

    def __enter__(self) -> Program:
        if self._strict:
            frame = inspect.currentframe()
            try:
                caller = frame.f_back if frame is not None else None
                if caller is not None:
                    _check_with_body_from_frame(caller, opt_out_hint="Program(strict=False)")
            finally:
                del frame
        Program._current = self
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        Program._current = None

    def add_rung(self, rung: RungLogic) -> None:
        """Add a rung to the program or current subroutine."""
        target = (
            self.subroutines[self._current_subroutine]
            if self._current_subroutine is not None
            else self.rungs
        )
        if rung._use_prior_snapshot and not target:
            scope = (
                f"subroutine '{self._current_subroutine}'"
                if self._current_subroutine is not None
                else "program"
            )
            raise RuntimeError(
                f"Rung.continued() cannot be the first rung in a {scope}. "
                "There is no prior condition snapshot to reuse."
            )
        target.append(rung)

    def start_subroutine(self, name: str) -> None:
        """Start defining a subroutine."""
        _validate_subroutine_name(name)
        self._current_subroutine = name
        self.subroutines[name] = []

    def end_subroutine(self) -> None:
        """End subroutine definition."""
        self._current_subroutine = None

    def call_subroutine(self, name: str, state: SystemState) -> SystemState:
        """Execute a subroutine by name (legacy state-based API)."""
        # This method is kept for backwards compatibility but should not be used
        # with ScanContext-based execution. Use call_subroutine_ctx instead.
        from pyrung.core.context import ScanContext

        if name not in self.subroutines:
            raise KeyError(f"Subroutine '{name}' not defined")
        ctx = ScanContext(state)
        for rung in self.subroutines[name]:
            rung.evaluate(ctx)
        return ctx.commit(dt=0.0)

    def call_subroutine_ctx(self, name: str, ctx: ScanContext) -> None:
        """Execute a subroutine by name within a ScanContext."""
        if name not in self.subroutines:
            raise KeyError(f"Subroutine '{name}' not defined")
        saved_snapshot = ctx._condition_snapshot
        ctx._condition_snapshot = None
        try:
            for rung in self.subroutines[name]:
                rung.evaluate(ctx)
        except SubroutineReturnSignal:
            pass
        finally:
            ctx._condition_snapshot = saved_snapshot

    @classmethod
    def current(cls) -> Program | None:
        """Get the current program context (if any)."""
        return cls._current

    @classmethod
    def register_dialect(cls, name: str, validator: DialectValidator) -> None:
        """Register a portability validator callback for a dialect name."""
        existing = cls._dialect_validators.get(name)
        if existing is None:
            cls._dialect_validators[name] = validator
            return
        if existing is validator:
            return
        raise ValueError(f"Dialect {name!r} already registered to a different validator")

    @classmethod
    def registered_dialects(cls) -> tuple[str, ...]:
        """Return registered dialect names in deterministic order."""
        return tuple(sorted(cls._dialect_validators))

    def validate(self, dialect: str, *, mode: str = "warn", **kwargs: Any) -> Any:
        """Run dialect-specific portability validation for this Program."""
        validator = self._dialect_validators.get(dialect)
        if validator is None:
            available = ", ".join(self.registered_dialects()) or "<none>"
            raise KeyError(
                f"Unknown validation dialect {dialect!r}. "
                f"Available dialects: {available}. "
                f"Import the dialect package first (example: import pyrung.{dialect})."
            )
        return validator(self, mode=mode, **kwargs)

    def evaluate(self, ctx: ScanContext) -> None:
        """Evaluate all main rungs in order (not subroutines) within a ScanContext."""
        for rung in self.rungs:
            rung.evaluate(ctx)


class Rung:
    """Context manager for defining a rung.

    Example:
        with Rung(Button):
            out(Light)

        with Rung(Step == 0):
            out(Light1)
            copy(1, Step, oneshot=True)
    """

    def __init__(self, *conditions: ConditionTerm) -> None:
        source_file, source_line = _capture_source(depth=2)
        self._rung = RungLogic(*conditions, source_file=source_file, source_line=source_line)
        self._pending_required_builder: tuple[int, str] | None = None
        condition_arg_lines = _capture_with_call_arg_lines(
            source_file,
            source_line,
            context_name="Rung",
        )

        # Consume any pending comment() call.
        prog = Program.current()
        if prog is not None and prog._pending_comment is not None:
            self._rung.comment = prog._pending_comment
            prog._pending_comment = None

        # Direct Tag conditions are converted internally and would otherwise
        # have no source metadata.
        for idx, condition in enumerate(self._rung._conditions):
            if condition.source_file is None:
                condition.source_file = source_file
            if condition.source_line is None:
                if idx < len(condition_arg_lines):
                    condition.source_line = condition_arg_lines[idx]
                else:
                    condition.source_line = source_line

    def continued(self) -> Rung:
        """Mark this rung as a continuation of the previous rung's condition snapshot.

        All conditions in this rung will evaluate against the same frozen
        state as the prior rung, rather than taking a fresh snapshot. This
        models multiple independent wires on the same visual rung in Click's
        ladder editor.

        Returns:
            self, for chaining: ``with Rung(B).continued(): ...``
        """
        if self._rung.comment is not None:
            raise RuntimeError(
                "A continued() rung cannot have its own comment. "
                "Set the comment on the original rung instead."
            )
        self._rung._use_prior_snapshot = True
        return self

    def _set_pending_required_builder(self, builder: object, descriptor: str) -> None:
        pending = self._pending_required_builder
        if pending is not None:
            _, existing = pending
            raise RuntimeError(f"{existing} must be completed before starting {descriptor}.")
        self._pending_required_builder = (id(builder), descriptor)

    def _assert_pending_required_builder_owner(self, builder: object, method_name: str) -> None:
        pending = self._pending_required_builder
        if pending is None:
            raise RuntimeError(
                f"{method_name}() called on a builder that is not pending in this flow."
            )
        pending_id, descriptor = pending
        if pending_id != id(builder):
            raise RuntimeError(
                f"{descriptor} must be completed before calling {method_name}() "
                "on a different builder."
            )

    def _clear_pending_required_builder(self, builder: object) -> None:
        self._assert_pending_required_builder_owner(builder, "finalize")
        self._pending_required_builder = None

    def _assert_no_pending_required_builder(self, next_action: str) -> None:
        pending = self._pending_required_builder
        if pending is None:
            return
        _, descriptor = pending
        raise RuntimeError(f"{descriptor} must be completed before calling {next_action}().")

    def _assert_required_builders_resolved(self, scope_name: str) -> None:
        pending = self._pending_required_builder
        if pending is None:
            return
        _, descriptor = pending
        raise RuntimeError(f"{descriptor} must be completed before exiting {scope_name} flow.")

    def __enter__(self) -> Rung:
        _push_rung_context(self)
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        try:
            if exc_type is None:
                self._assert_required_builders_resolved("Rung")
        finally:
            _pop_rung_context()
        _set_scope_end_line(
            self._rung,
            source_file=self._rung.source_file,
            source_line=self._rung.source_line,
            context_name="Rung",
        )
        # Add rung to current program
        prog = Program.current()
        if prog is not None:
            prog.add_rung(self._rung)


class Subroutine:
    """Context manager for defining a subroutine.

    Subroutines are named blocks of rungs that are only executed when called.

    Example:
        with subroutine("my_sub"):
            with Rung():
                out(Light)
    """

    def __init__(self, name: str, *, strict: bool = True) -> None:
        self._name = _validate_subroutine_name(name)
        self._strict = strict

    def __enter__(self) -> Subroutine:
        prog = Program.current()
        if prog is None:
            raise RuntimeError("subroutine() must be used inside a Program context")
        prog.start_subroutine(self._name)
        return self

    def __call__(self, fn: Callable[[], None]) -> SubroutineFunc:
        """Use subroutine() as a decorator.

        Example:
            @subroutine("init")
            def init_sequence():
                with Rung():
                    out(Light)
        """
        if self._strict:
            _check_function_body_strict(
                fn,
                opt_out_hint=f'@subroutine("{self._name}", strict=False)',
                source_label=f'@subroutine("{self._name}") {getattr(fn, "__qualname__", repr(fn))}',
            )
        return SubroutineFunc(self._name, fn)

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        prog = Program.current()
        if prog is not None:
            prog.end_subroutine()


class SubroutineFunc:
    """A decorated function that represents a subroutine.

    Created by using @subroutine("name") as a decorator. When passed to call(),
    auto-registers with the current Program on first use.

    Example:
        @subroutine("init")
        def init_sequence():
            with Rung():
                out(Light)

        with Program() as logic:
            with Rung(Button):
                call(init_sequence)
    """

    def __init__(self, name: str, fn: Callable[[], None]) -> None:
        self._name = name
        self._fn = fn

    @property
    def name(self) -> str:
        """The subroutine name."""
        return self._name

    def _register(self, prog: Program) -> None:
        """Register this subroutine's rungs with a Program."""
        prog.start_subroutine(self._name)
        self._fn()
        prog.end_subroutine()


def subroutine(name: str, *, strict: bool = True) -> Subroutine:
    """Define a named subroutine.

    Subroutines are only executed when called via call().
    They are NOT executed during normal program scan.

    Example:
        with Program() as logic:
            with Rung(Button):
                call("my_sub")

            with subroutine("my_sub"):
                with Rung():
                    out(Light)
    """
    return Subroutine(name, strict=strict)


class ForLoop:
    """Context manager for a repeated instruction block within a rung."""

    def __init__(
        self,
        count: Tag | int,
        oneshot: bool = False,
        source_file: str | None = None,
        source_line: int | None = None,
    ) -> None:
        self.count = count
        self.oneshot = oneshot
        self.idx = Tag("_forloop_idx", TagType.DINT)
        self._parent_ctx: Rung | None = None
        self._capture_rung: RungLogic | None = None
        self._capture_ctx: Rung | None = None
        self._source_file = source_file
        self._source_line = source_line

    def __enter__(self) -> ForLoop:
        global _forloop_active

        if _forloop_active:
            raise RuntimeError("Nested forloop is not permitted")

        self._parent_ctx = _require_rung_context("forloop")
        self._parent_ctx._assert_no_pending_required_builder("forloop")
        _forloop_active = True

        # Capture body instructions to a temporary rung (like Branch capture).
        self._capture_rung = RungLogic(
            source_file=self._source_file,
            source_line=self._source_line,
        )
        self._capture_ctx = _new_capture_context(self._capture_rung)
        _push_rung_context(self._capture_ctx)

        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        global _forloop_active

        _pop_rung_context()
        _forloop_active = False

        if exc_type is None and self._capture_ctx is not None:
            self._capture_ctx._assert_required_builders_resolved("forloop")

        if self._parent_ctx is None or self._capture_rung is None:
            return

        _set_scope_end_line(
            self._capture_rung,
            source_file=self._source_file,
            source_line=self._source_line,
            context_name="forloop",
        )

        instruction = ForLoopInstruction(
            count=self.count,
            idx_tag=self.idx,
            instructions=self._capture_rung._instructions,
            oneshot=self.oneshot,
        )
        instruction.source_file, instruction.source_line = self._source_file, self._source_line
        self._parent_ctx._rung.add_instruction(instruction)


def forloop(count: Tag | int, oneshot: bool = False) -> ForLoop:
    """Create a repeated instruction block context.

    Example:
        with Rung(Enable):
            with forloop(10) as loop:
                copy(Source[loop.idx + 1], Dest[loop.idx + 1])
    """
    source_file, source_line = _capture_source(depth=2)
    return ForLoop(count, oneshot=oneshot, source_file=source_file, source_line=source_line)


class Branch:
    """Context manager for a parallel branch within a rung.

    A branch executes when both the parent rung conditions AND
    the branch's own conditions are true.

    Example:
        with Rung(Step == 0):
            out(Light1)
            with branch(AutoMode):  # Only executes if Step==0 AND AutoMode
                out(Light2)
                copy(1, Step, oneshot=True)
    """

    def __init__(
        self,
        *conditions: ConditionTerm,
        source_file: str | None = None,
        source_line: int | None = None,
    ) -> None:
        """Create a branch with additional conditions.

        Args:
            conditions: Conditions that must be true (in addition to parent rung)
                        for this branch's instructions to execute.
        """
        self._conditions = list(conditions)
        self._branch_rung: RungLogic | None = None
        self._parent_ctx: Rung | None = None
        self._branch_ctx: Rung | None = None
        self._source_file = source_file
        self._source_line = source_line

    def __enter__(self) -> Branch:
        if _forloop_active:
            raise RuntimeError("branch() is not permitted inside forloop()")

        # Get parent rung context
        self._parent_ctx = _current_rung()
        if self._parent_ctx is None:
            raise RuntimeError("branch() must be called inside a Rung context")
        self._parent_ctx._assert_no_pending_required_builder("branch")

        # Create a nested rung for the branch that includes BOTH parent and branch conditions
        # This ensures terminal instructions (counters, timers) see the full condition chain
        parent_conditions = self._parent_ctx._rung._conditions
        condition_arg_lines = _capture_with_call_arg_lines(
            self._source_file,
            self._source_line,
            context_name="branch",
        )
        combined_conditions = parent_conditions + self._conditions
        self._branch_rung = RungLogic(
            *combined_conditions,
            source_file=self._source_file,
            source_line=self._source_line,
        )
        self._branch_rung._branch_condition_start = len(parent_conditions)

        local_conditions = self._branch_rung._conditions[
            self._branch_rung._branch_condition_start :
        ]
        for idx, condition in enumerate(local_conditions):
            if condition.source_line is None:
                if idx < len(condition_arg_lines):
                    condition.source_line = condition_arg_lines[idx]
                else:
                    condition.source_line = self._source_line

        for condition in self._branch_rung._conditions:
            if condition.source_file is None:
                condition.source_file = self._source_file
            if condition.source_line is None:
                condition.source_line = self._source_line

        # Push a new "fake" rung context so instructions go to the branch
        self._branch_ctx = _new_capture_context(self._branch_rung)
        _push_rung_context(self._branch_ctx)

        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        # Pop our branch context
        _pop_rung_context()

        if exc_type is None and self._branch_ctx is not None:
            self._branch_ctx._assert_required_builders_resolved("branch")

        # Add the branch as a nested rung to the parent
        if self._parent_ctx is not None and self._branch_rung is not None:
            _set_scope_end_line(
                self._branch_rung,
                source_file=self._source_file,
                source_line=self._source_line,
                context_name="branch",
            )
            self._parent_ctx._rung.add_branch(self._branch_rung)


def branch(*conditions: ConditionTerm) -> Branch:
    """Create a parallel branch within a rung.

    A branch executes when both the parent rung conditions AND
    the branch's own conditions are true.

    Example:
        with Rung(Step == 0):
            out(Light1)
            with branch(AutoMode):  # Only executes if Step==0 AND AutoMode
                out(Light2)
                copy(1, Step, oneshot=True)

    Args:
        conditions: Conditions that must be true (in addition to parent rung)
                    for this branch's instructions to execute.

    Returns:
        Branch context manager.
    """
    source_file, source_line = _capture_source(depth=2)
    return Branch(*conditions, source_file=source_file, source_line=source_line)


_MAX_COMMENT_LENGTH = 1400


def comment(text: str) -> None:
    """Attach a comment to the next rung.

    The comment is consumed by the immediately following ``Rung()`` constructor.
    Must be called inside a ``Program`` context.

    Example::

        comment("UnitMode Change")
        with Rung(C_UnitModeChgRequest):
            copy(1, C_UnitModeChgRequestBool, oneshot=True)
    """
    prog = Program.current()
    if prog is None:
        raise RuntimeError("comment() must be used inside a Program context")
    if prog._pending_comment is not None:
        raise RuntimeError("comment() already set — missing a Rung after the previous comment()?")
    text = textwrap.dedent(text).strip()
    if len(text) > _MAX_COMMENT_LENGTH:
        raise ValueError(f"Rung comment is {len(text)} chars, max is {_MAX_COMMENT_LENGTH}.")
    prog._pending_comment = text


RungContext = Rung



"""Tests for source-file/line metadata capture on DSL elements."""

from __future__ import annotations

import inspect
from pathlib import Path
from types import FrameType
from typing import cast

from pyrung.core import (
    Block,
    Bool,
    Int,
    PLCRunner,
    Program,
    Rung,
    TagType,
    all_of,
    any_of,
    branch,
    copy,
    count_down,
    count_up,
    event_drum,
    forloop,
    latch,
    on_delay,
    out,
    rise,
    shift,
    time_drum,
)
from pyrung.core.instruction import ForLoopInstruction


def _line_no() -> int:
    frame = cast(FrameType, inspect.currentframe())
    caller = cast(FrameType, frame.f_back)
    return caller.f_lineno


def test_rung_captures_source_file_start_line_and_end_line():
    enable = Bool("Enable")
    light = Bool("Light")
    latched = Bool("Latched")

    with Program(strict=False) as prog:
        rung_line = _line_no() + 1
        with Rung(enable):
            out_line = _line_no() + 1
            out(light)
            end_line = _line_no() + 1
            latch(latched)

    rung = prog.rungs[0]
    assert rung.source_file is not None
    assert Path(rung.source_file).name == Path(__file__).name
    assert rung.source_line == rung_line
    assert rung.end_line == end_line
    assert rung._instructions[0].source_line == out_line


def test_condition_helpers_and_operators_capture_source_lines():
    a = Bool("A")
    b = Bool("B")
    c = Bool("C")
    step = Int("Step")

    line_nc = _line_no() + 1
    cond_nc = ~a
    line_rise = _line_no() + 1
    cond_rise = rise(b)
    line_any = _line_no() + 1
    cond_any = any_of(a, b)
    line_all = _line_no() + 1
    cond_all = all_of(a, c)
    line_or = _line_no() + 1
    cond_or = a | b
    line_and = _line_no() + 1
    cond_and = a & b
    line_cmp = _line_no() + 1
    cond_cmp = step == 10
    line_expr_cmp = _line_no() + 1
    cond_expr_cmp = (step + 1) > 0
    line_chain = _line_no() + 1
    cond_chain = (a | b) | (step > 1)

    assert cond_nc.source_line == line_nc
    assert cond_rise.source_line == line_rise
    assert cond_any.source_line == line_any
    assert cond_all.source_line == line_all
    assert cond_or.source_line == line_or
    assert cond_and.source_line == line_and
    assert cond_cmp.source_line == line_cmp
    assert cond_expr_cmp.source_line == line_expr_cmp
    assert cond_chain.source_line == line_chain

    # Direct-Tag children get inherited source metadata when they are wrapped.
    assert cond_or.conditions[0].source_line == line_or
    assert cond_or.conditions[1].source_line == line_or


def test_builder_paths_capture_source_lines_for_branch_forloop_and_terminal_instructions():
    enable = Bool("Enable")
    branch_enable = Bool("BranchEnable")
    branch_out = Bool("BranchOut")
    forloop_counter = Int("ForLoopCounter")
    reset_cond = Bool("Reset")
    cu_done = Bool("CountUpDone")
    cu_acc = Int("CountUpAcc")
    cd_done = Bool("CountDownDone")
    cd_acc = Int("CountDownAcc")
    timer_done = Bool("TimerDone")
    timer_acc = Int("TimerAcc")
    drum_step = Int("DrumStep")
    drum_acc = Int("DrumAcc")
    drum_done = Bool("DrumDone")
    drum_out1 = Bool("DrumOut1")
    drum_out2 = Bool("DrumOut2")
    drum_event1 = Bool("DrumEvent1")
    drum_event2 = Bool("DrumEvent2")
    clock = Bool("Clock")
    bits = Block("C", TagType.BOOL, 1, 8)

    with Program(strict=False) as prog:
        with Rung(enable):
            branch_line = _line_no() + 1
            with branch(branch_enable):
                branch_end_line = _line_no() + 1
                out(branch_out)

        with Rung(enable):
            forloop_line = _line_no() + 1
            with forloop(2):
                forloop_copy_line = _line_no() + 1
                copy(forloop_counter + 1, forloop_counter)

        with Rung(enable):
            count_up_line = _line_no() + 1
            count_up(cu_done, cu_acc, preset=5).reset(reset_cond)

        with Rung(enable):
            count_down_line = _line_no() + 1
            count_down(cd_done, cd_acc, preset=5).reset(reset_cond)

        with Rung(enable):
            on_delay_line = _line_no() + 1
            on_delay(timer_done, timer_acc, preset=50).reset(reset_cond)

        with Rung(enable):
            shift_line = _line_no() + 1
            shift(bits.select(1, 4)).clock(clock).reset(reset_cond)

        with Rung(enable):
            event_drum_line = _line_no() + 1
            event_drum(
                outputs=[drum_out1, drum_out2],
                events=[drum_event1, drum_event2],
                pattern=[[1, 0], [0, 1]],
                current_step=drum_step,
                completion_flag=drum_done,
            ).reset(reset_cond)

        with Rung(enable):
            time_drum_line = _line_no() + 1
            time_drum(
                outputs=[drum_out1, drum_out2],
                presets=[100, 100],
                pattern=[[1, 0], [0, 1]],
                current_step=drum_step,
                accumulator=drum_acc,
                completion_flag=drum_done,
            ).reset(reset_cond)

    branch_rung = prog.rungs[0]._branches[0]
    assert branch_rung.source_line == branch_line
    assert branch_rung.end_line == branch_end_line

    forloop_instr = prog.rungs[1]._instructions[0]
    assert isinstance(forloop_instr, ForLoopInstruction)
    assert forloop_instr.source_line == forloop_line
    assert forloop_instr.instructions[0].source_line == forloop_copy_line

    count_up_instr = prog.rungs[2]._instructions[0]
    assert type(count_up_instr).__name__ == "CountUpInstruction"
    assert count_up_instr.source_line == count_up_line

    count_down_instr = prog.rungs[3]._instructions[0]
    assert type(count_down_instr).__name__ == "CountDownInstruction"
    assert count_down_instr.source_line == count_down_line

    on_delay_instr = prog.rungs[4]._instructions[0]
    assert type(on_delay_instr).__name__ == "OnDelayInstruction"
    assert on_delay_instr.source_line == on_delay_line

    shift_instr = prog.rungs[5]._instructions[0]
    assert type(shift_instr).__name__ == "ShiftInstruction"
    assert shift_instr.source_line == shift_line

    event_drum_instr = prog.rungs[6]._instructions[0]
    assert type(event_drum_instr).__name__ == "EventDrumInstruction"
    assert event_drum_instr.source_line == event_drum_line

    time_drum_instr = prog.rungs[7]._instructions[0]
    assert type(time_drum_instr).__name__ == "TimeDrumInstruction"
    assert time_drum_instr.source_line == time_drum_line


def test_multiline_rung_direct_tag_condition_uses_argument_line_number():
    step = Int("Step")
    auto_mode = Bool("AutoMode")
    light = Bool("Light")

    with Program(strict=False) as prog:
        marker_line = _line_no()
        with Rung(
            step == 1,
            auto_mode,
        ):
            out(light)

    step_condition_line = marker_line + 2
    auto_condition_line = marker_line + 3
    rung = prog.rungs[0]
    assert rung._conditions[0].source_line == step_condition_line
    assert rung._conditions[1].source_line == auto_condition_line


def test_multiline_count_up_captures_instruction_end_line_and_debug_step_end_line():
    enable = Bool("Enable")
    done = Bool("Done")
    acc = Int("Acc")
    reset_cond = Bool("Reset")

    with Program(strict=False) as prog:
        with Rung(enable):
            count_up_line = _line_no() + 1
            count_up(
                done,
                acc,
                preset=5,
            ).reset(reset_cond)
            count_up_end_line = _line_no() - 1

    instruction = prog.rungs[0]._instructions[0]
    assert instruction.source_line == count_up_line
    assert instruction.end_line == count_up_end_line

    runner = PLCRunner(prog)
    runner.patch({"Enable": True, "Reset": False})
    scan_gen = runner.scan_steps_debug()
    rung_step = next(scan_gen)
    assert rung_step.kind == "rung"
    step = next(scan_gen)
    assert step.kind == "instruction"
    assert step.source_line == count_up_line
    assert step.end_line == count_up_line


def test_chained_builder_methods_capture_distinct_debug_substep_lines():
    enable = Bool("Enable")
    down = Bool("Down")
    reset_cond = Bool("Reset")
    clock = Bool("Clock")
    up_done = Bool("UpDone")
    up_acc = Int("UpAcc")
    down_done = Bool("DownDone")
    down_acc = Int("DownAcc")
    timer_done = Bool("TimerDone")
    timer_acc = Int("TimerAcc")
    drum_step = Int("DrumStep")
    drum_acc = Int("DrumAcc")
    drum_done = Bool("DrumDone")
    drum_out1 = Bool("DrumOut1")
    drum_out2 = Bool("DrumOut2")
    drum_event1 = Bool("DrumEvent1")
    drum_event2 = Bool("DrumEvent2")
    jump = Bool("Jump")
    jog = Bool("Jog")
    bits = Block("C", TagType.BOOL, 1, 8)

    with Program(strict=False) as prog:
        with Rung(enable):
            cu_line = _line_no() + 1
            cu_builder = count_up(up_done, up_acc, preset=5)
            cu_down_line = _line_no() + 1
            cu_builder = cu_builder.down(down)
            cu_reset_line = _line_no() + 1
            cu_builder.reset(reset_cond)

        with Rung(enable):
            cd_line = _line_no() + 1
            cd_builder = count_down(down_done, down_acc, preset=5)
            cd_reset_line = _line_no() + 1
            cd_builder.reset(reset_cond)

        with Rung(enable):
            timer_line = _line_no() + 1
            timer_builder = on_delay(timer_done, timer_acc, preset=50)
            timer_reset_line = _line_no() + 1
            timer_builder.reset(reset_cond)

        with Rung(enable):
            shift_line = _line_no() + 1
            shift_builder = shift(bits.select(1, 4))
            shift_clock_line = _line_no() + 1
            shift_builder = shift_builder.clock(clock)
            shift_reset_line = _line_no() + 1
            shift_builder.reset(reset_cond)

        with Rung(enable):
            event_line = _line_no() + 1
            event_builder = event_drum(
                outputs=[drum_out1, drum_out2],
                events=[drum_event1, drum_event2],
                pattern=[[1, 0], [0, 1]],
                current_step=drum_step,
                completion_flag=drum_done,
            )
            event_reset_line = _line_no() + 1
            event_builder = event_builder.reset(reset_cond)
            event_jump_line = _line_no() + 1
            event_builder = event_builder.jump(jump, step=drum_step)
            event_jog_line = _line_no() + 1
            event_builder.jog(jog)

        with Rung(enable):
            time_line = _line_no() + 1
            time_builder = time_drum(
                outputs=[drum_out1, drum_out2],
                presets=[100, 100],
                pattern=[[1, 0], [0, 1]],
                current_step=drum_step,
                accumulator=drum_acc,
                completion_flag=drum_done,
            )
            time_reset_line = _line_no() + 1
            time_builder = time_builder.reset(reset_cond)
            time_jump_line = _line_no() + 1
            time_builder = time_builder.jump(jump, step=drum_step)
            time_jog_line = _line_no() + 1
            time_builder.jog(jog)

    cu_instr = prog.rungs[0]._instructions[0]
    assert cu_instr.debug_substeps is not None
    assert [step.instruction_kind for step in cu_instr.debug_substeps] == [
        "Count Up",
        "Count Down",
        "Reset",
    ]
    assert [step.source_line for step in cu_instr.debug_substeps] == [
        cu_line,
        cu_down_line,
        cu_reset_line,
    ]

    cd_instr = prog.rungs[1]._instructions[0]
    assert cd_instr.debug_substeps is not None
    assert [step.instruction_kind for step in cd_instr.debug_substeps] == [
        "Count Down",
        "Reset",
    ]
    assert [step.source_line for step in cd_instr.debug_substeps] == [
        cd_line,
        cd_reset_line,
    ]

    timer_instr = prog.rungs[2]._instructions[0]
    assert timer_instr.debug_substeps is not None
    assert [step.instruction_kind for step in timer_instr.debug_substeps] == [
        "Enable",
        "Reset",
    ]
    assert [step.source_line for step in timer_instr.debug_substeps] == [
        timer_line,
        timer_reset_line,
    ]

    shift_instr = prog.rungs[3]._instructions[0]
    assert shift_instr.debug_substeps is not None
    assert [step.instruction_kind for step in shift_instr.debug_substeps] == [
        "Data",
        "Clock",
        "Reset",
    ]
    assert [step.source_line for step in shift_instr.debug_substeps] == [
        shift_line,
        shift_clock_line,
        shift_reset_line,
    ]

    event_instr = prog.rungs[4]._instructions[0]
    assert event_instr.debug_substeps is not None
    assert [step.instruction_kind for step in event_instr.debug_substeps] == [
        "Auto",
        "Reset",
        "Jump",
        "Jog",
    ]
    assert [step.source_line for step in event_instr.debug_substeps] == [
        event_line,
        event_reset_line,
        event_jump_line,
        event_jog_line,
    ]

    time_instr = prog.rungs[5]._instructions[0]
    assert time_instr.debug_substeps is not None
    assert [step.instruction_kind for step in time_instr.debug_substeps] == [
        "Auto",
        "Reset",
        "Jump",
        "Jog",
    ]
    assert [step.source_line for step in time_instr.debug_substeps] == [
        time_line,
        time_reset_line,
        time_jump_line,
        time_jog_line,
    ]

    runner = PLCRunner(prog)
    runner.patch({"Enable": True, "Down": True, "Reset": False, "Clock": True})
    instruction_steps = [step for step in runner.scan_steps_debug() if step.kind == "instruction"]
    assert [step.source_line for step in instruction_steps] == [
        cu_line,
        cu_down_line,
        cu_reset_line,
        cd_line,
        cd_reset_line,
        timer_line,
        timer_reset_line,
        shift_line,
        shift_clock_line,
        shift_reset_line,
        event_line,
        event_reset_line,
        event_jump_line,
        event_jog_line,
        time_line,
        time_reset_line,
        time_jump_line,
        time_jog_line,
    ]



"""Source-location capture helpers for DSL metadata."""

from __future__ import annotations

import ast
import inspect
from functools import lru_cache
from pathlib import Path


def _capture_source(depth: int = 2) -> tuple[str | None, int | None]:
    """Capture (filename, line number) from the caller stack."""
    frame = inspect.currentframe()
    try:
        for _ in range(depth):
            if frame is None:
                return (None, None)
            frame = frame.f_back
        if frame is None:
            return (None, None)
        return (frame.f_code.co_filename, frame.f_lineno)
    finally:
        del frame


def _capture_with_end_line(
    source_file: str | None,
    start_line: int | None,
    *,
    context_name: str | None = None,
) -> int | None:
    """Best-effort lookup of a with-statement end line from source."""
    if source_file is None or start_line is None:
        return None
    if source_file.startswith("<"):
        return None

    tree = _parse_file_ast(source_file)
    if tree is None:
        return None

    candidates: list[tuple[int, int]] = []
    for node in ast.walk(tree):
        if not isinstance(node, (ast.With, ast.AsyncWith)):
            continue
        if getattr(node, "lineno", None) != start_line:
            continue

        end_line = getattr(node, "end_lineno", None)
        if end_line is None:
            continue

        if context_name is None:
            candidates.append((0, end_line))
            continue

        score = 1
        for item in node.items:
            name = _context_expr_name(item.context_expr)
            if name == context_name:
                score = 0
                break
        candidates.append((score, end_line))

    if not candidates:
        return None

    candidates.sort(key=lambda pair: pair[0])
    return candidates[0][1]


def _capture_with_call_arg_lines(
    source_file: str | None,
    start_line: int | None,
    *,
    context_name: str | None = None,
) -> list[int]:
    """Best-effort lookup of `with` context-manager call argument line numbers."""
    if source_file is None or start_line is None:
        return []
    if source_file.startswith("<"):
        return []

    tree = _parse_file_ast(source_file)
    if tree is None:
        return []

    candidates: list[tuple[int, list[int]]] = []
    for node in ast.walk(tree):
        if not isinstance(node, (ast.With, ast.AsyncWith)):
            continue
        if getattr(node, "lineno", None) != start_line:
            continue

        for item in node.items:
            context_expr = item.context_expr
            if not isinstance(context_expr, ast.Call):
                continue

            if context_name is None:
                score = 0
            else:
                name = _context_expr_name(context_expr.func)
                score = 0 if name == context_name else 1

            arg_lines: list[int] = []
            for arg in context_expr.args:
                line = getattr(arg, "lineno", None)
                if line is None:
                    continue
                arg_lines.append(int(line))
            candidates.append((score, arg_lines))

    if not candidates:
        return []

    candidates.sort(key=lambda pair: pair[0])
    return candidates[0][1]


def _capture_call_end_line(
    source_file: str | None,
    start_line: int | None,
    *,
    call_name: str | None = None,
) -> int | None:
    """Best-effort lookup of a call-expression end line from source."""
    if source_file is None or start_line is None:
        return None
    if source_file.startswith("<"):
        return None

    tree = _parse_file_ast(source_file)
    if tree is None:
        return None

    candidates: list[tuple[int, int]] = []
    for node in ast.walk(tree):
        if not isinstance(node, ast.Call):
            continue
        if getattr(node, "lineno", None) != start_line:
            continue

        end_line = getattr(node, "end_lineno", None)
        if end_line is None:
            continue

        if call_name is None:
            score = 0
        else:
            name = _context_expr_name(node.func)
            score = 0 if name == call_name else 1

        candidates.append((score, int(end_line)))

    if not candidates:
        return None

    candidates.sort(key=lambda pair: (pair[0], -pair[1]))
    return candidates[0][1]


@lru_cache(maxsize=256)
def _parse_file_ast(source_file: str) -> ast.AST | None:
    try:
        source = Path(source_file).read_text(encoding="utf-8")
    except OSError:
        return None
    try:
        return ast.parse(source, filename=source_file)
    except SyntaxError:
        return None


def _context_expr_name(expr: ast.AST) -> str | None:
    if isinstance(expr, ast.Call):
        return _context_expr_name(expr.func)
    if isinstance(expr, ast.Name):
        return expr.id
    if isinstance(expr, ast.Attribute):
        return expr.attr
    return None



"""Tests for table-driven hardware capabilities."""

from __future__ import annotations

import pytest

from pyrung.click.capabilities import (
    CLICK_HARDWARE_PROFILE,
    COMPARE_COMPATIBILITY,
    COMPARE_CONSTANT_COMPATIBILITY,
    COPY_COMPATIBILITY,
    INSTRUCTION_ROLE_COMPATIBILITY,
    LADDER_WRITABLE_SC,
    LADDER_WRITABLE_SD,
    ClickHardwareProfile,
    CompareConstantKind,
    CopyOperation,
    InstructionRole,
)


def test_profile_export_available():
    assert isinstance(CLICK_HARDWARE_PROFILE, ClickHardwareProfile)


@pytest.mark.parametrize(
    ("memory_type", "address", "expected"),
    [
        ("X", 1, False),
        ("Y", 1, True),
        ("C", 1, True),
        ("T", 1, False),
        ("CT", 1, False),
        ("DS", 1, True),
        ("DD", 1, True),
        ("DH", 1, True),
        ("DF", 1, True),
        ("XD", 1, False),
        ("YD", 1, False),
        ("TD", 1, True),
        ("CTD", 1, True),
        ("TXT", 1, True),
    ],
)
def test_is_writable_baseline(memory_type: str, address: int, expected: bool):
    assert CLICK_HARDWARE_PROFILE.is_writable(memory_type, address) is expected


def test_is_writable_sc_subset():
    for address in LADDER_WRITABLE_SC:
        assert CLICK_HARDWARE_PROFILE.is_writable("SC", address) is True
    assert CLICK_HARDWARE_PROFILE.is_writable("SC", 1) is False
    assert CLICK_HARDWARE_PROFILE.is_writable("SC", None) is False


def test_sd_control_sc_addresses_remain_writable():
    for address in (65, 66, 67):
        assert address in LADDER_WRITABLE_SC
        assert CLICK_HARDWARE_PROFILE.is_writable("SC", address) is True


def test_is_writable_sd_subset():
    for address in LADDER_WRITABLE_SD:
        assert CLICK_HARDWARE_PROFILE.is_writable("SD", address) is True
    assert CLICK_HARDWARE_PROFILE.is_writable("SD", 1) is False
    assert CLICK_HARDWARE_PROFILE.is_writable("SD", None) is False


@pytest.mark.parametrize(
    ("role", "ok_bank", "bad_bank"),
    [
        ("timer_done_bit", "T", "C"),
        ("timer_accumulator", "TD", "DS"),
        ("timer_preset", "DS", "DD"),
        ("counter_done_bit", "CT", "C"),
        ("counter_accumulator", "CTD", "DD"),
        ("counter_preset", "DD", "TD"),
        ("drum_output_bit", "Y", "DS"),
        ("drum_current_step", "DS", "DD"),
        ("drum_completion_flag", "C", "Y"),
        ("drum_accumulator", "TD", "DS"),
        ("drum_jump_step", "DS", "DD"),
        ("drum_event_condition", "X", "DS"),
        ("copy_pointer", "DS", "DD"),
    ],
)
def test_role_compatibility(role: InstructionRole, ok_bank: str, bad_bank: str):
    assert CLICK_HARDWARE_PROFILE.valid_for_role(ok_bank, role) is True
    assert CLICK_HARDWARE_PROFILE.valid_for_role(bad_bank, role) is False


@pytest.mark.parametrize(
    ("operation", "source", "dest", "expected"),
    [
        ("single", "X", "Y", True),
        ("single", "DS", "TXT", True),
        ("single", "DS", "C", False),
        ("block", "TXT", "DS", True),
        ("block", "TXT", "TXT", False),
        ("fill", "TXT", "TXT", True),
        ("fill", "TXT", "DS", False),
        ("pack_bits", "X", "DS", True),
        ("pack_bits", "C", "DF", True),
        ("pack_bits", "X", "DD", False),
        ("pack_words", "DS", "DD", True),
        ("pack_words", "DD", "DF", False),
        ("unpack_bits", "DD", "Y", True),
        ("unpack_bits", "DD", "DS", False),
        ("unpack_words", "DF", "DH", True),
        ("unpack_words", "DS", "DH", False),
    ],
)
def test_copy_compatibility(operation: CopyOperation, source: str, dest: str, expected: bool):
    assert CLICK_HARDWARE_PROFILE.copy_compatible(operation, source, dest) is expected


@pytest.mark.parametrize(
    ("left", "right", "expected"),
    [
        ("DS", "DD", True),
        ("DH", "XD", True),
        ("TXT", "TXT", True),
        ("TXT", "DD", False),
        ("DH", "DD", False),
    ],
)
def test_compare_compatibility(left: str, right: str, expected: bool):
    assert CLICK_HARDWARE_PROFILE.compare_compatible(left, right) is expected


@pytest.mark.parametrize(
    ("bank", "const_kind", "expected"),
    [
        ("DS", "int1", True),
        ("DS", "float", True),
        ("DH", "hex", True),
        ("TXT", "text", True),
        ("DH", "int1", False),
        ("TXT", "hex", False),
    ],
)
def test_compare_constant_compatibility(bank: str, const_kind: CompareConstantKind, expected: bool):
    assert CLICK_HARDWARE_PROFILE.compare_constant_compatible(bank, const_kind) is expected


def test_lookup_tables_exported():
    assert "single" in COPY_COMPATIBILITY
    assert "timer_done_bit" in INSTRUCTION_ROLE_COMPATIBILITY
    assert "drum_output_bit" in INSTRUCTION_ROLE_COMPATIBILITY
    assert ("DS", "DD") in COMPARE_COMPATIBILITY
    assert "DS" in COMPARE_CONSTANT_COMPATIBILITY


@pytest.mark.parametrize(
    "call",
    [
        lambda: CLICK_HARDWARE_PROFILE.is_writable("ZZ", 1),
        lambda: CLICK_HARDWARE_PROFILE.valid_for_role("ZZ", "timer_done_bit"),
        lambda: CLICK_HARDWARE_PROFILE.valid_for_role("T", "bad_role"),  # ty: ignore[invalid-argument-type]
        lambda: CLICK_HARDWARE_PROFILE.copy_compatible("bad_op", "DS", "DD"),  # ty: ignore[invalid-argument-type]
        lambda: CLICK_HARDWARE_PROFILE.copy_compatible("single", "ZZ", "DD"),
        lambda: CLICK_HARDWARE_PROFILE.copy_compatible("single", "DS", "ZZ"),
        lambda: CLICK_HARDWARE_PROFILE.compare_compatible("ZZ", "DS"),
        lambda: CLICK_HARDWARE_PROFILE.compare_constant_compatible("ZZ", "int1"),
        lambda: CLICK_HARDWARE_PROFILE.compare_constant_compatible("DS", "bad_kind"),  # ty: ignore[invalid-argument-type]
    ],
)
def test_unknown_inputs_raise_key_error(call):
    with pytest.raises(KeyError):
        call()



"""Dump wild ladder patterns as CSV for visual review."""

from __future__ import annotations

import io
import sys
import tempfile
from pathlib import Path

sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding="utf-8")

from pyrung.click import TagMap, c, ct, ctd, pyrung_to_ladder, x, y
from pyrung.core import Bool, Dint, Program, Rung, any_of
from pyrung.core.program import all_of, branch, count_up, latch, out

OUT = Path(tempfile.mkdtemp(prefix="ladder_"))


def _dump(label: str, bundle):
    safe = label.replace(" ", "_").replace(",", "").replace(":", "")[:60]
    d = OUT / safe
    d.mkdir(parents=True, exist_ok=True)
    bundle.write(d)
    print(f"\n== {label} ==")
    print((d / "main.csv").read_text(encoding="utf-8"), end="")


# --- Tags ---
A = Bool("A")
B = Bool("B")
C = Bool("C")
D = Bool("D")
E = Bool("E")
F = Bool("F")
G = Bool("G")
Mode = Bool("Mode")
Mode2 = Bool("Mode2")
Y1 = Bool("Y1")
Y2 = Bool("Y2")
Y3 = Bool("Y3")
Y4 = Bool("Y4")

base_map = {
    A: x[1],
    B: x[2],
    C: x[3],
    D: x[4],
    E: x[5],
    F: x[6],
    G: x[7],
    Mode: c[1],
    Mode2: c[2],
    Y1: y[1],
    Y2: y[2],
    Y3: y[3],
    Y4: y[4],
}

# ── Case 1: Deep AND chain + branch ──
with Program() as logic:
    with Rung(A, B, C, D):
        out(Y1)
        with branch(E):
            out(Y2)
        with branch(F):
            out(Y3)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 1: A,B,C,D + branch(E) + branch(F)", pyrung_to_ladder(logic, mapping))

# ── Case 2: Deep AND + branch with its own condition ──
with Program() as logic:
    with Rung(A, B, C):
        out(Y1)
        with branch(D, E):
            out(Y2)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 2: A,B,C + branch(D,E)", pyrung_to_ladder(logic, mapping))

# ── Case 3: Single condition + many branches ──
with Program() as logic:
    with Rung(A):
        out(Y1)
        with branch(B):
            out(Y2)
        with branch(C):
            out(Y3)
        with branch(D):
            out(Y4)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 3: A + branch(B) + branch(C) + branch(D)", pyrung_to_ladder(logic, mapping))

# ── Case 4: Branch after multiple outputs ──
with Program() as logic:
    with Rung(A):
        out(Y1)
        latch(Y2)
        with branch(Mode):
            out(Y3)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 4: A out+latch + branch(Mode)", pyrung_to_ladder(logic, mapping))

# ── Case 5: Branch sandwiched between instructions ──
with Program() as logic:
    with Rung(A, B):
        out(Y1)
        with branch(Mode):
            out(Y2)
        out(Y3)
        with branch(Mode2):
            out(Y4)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 5: A,B out branch out branch", pyrung_to_ladder(logic, mapping))

# ── Case 6: any_of inside all_of ──
with Program() as logic:
    with Rung(all_of(A, any_of(B, C), D)):
        out(Y1)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 6: all_of(A, any_of(B,C), D)", pyrung_to_ladder(logic, mapping))

# ── Case 7: any_of(all_of, all_of) ──
with Program() as logic:
    with Rung(any_of(all_of(A, B), all_of(C, D))):
        out(Y1)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 7: any_of(all_of(A,B), all_of(C,D))", pyrung_to_ladder(logic, mapping))

# ── Case 8: any_of with mixed leaf/chain ──
with Program() as logic:
    with Rung(any_of(A, B, C, all_of(D, E, F))):
        out(Y1)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 8: any_of(A, B, C, all_of(D,E,F))", pyrung_to_ladder(logic, mapping))

# ── Case 9: Deep AND then OR ──
with Program() as logic:
    with Rung(A, B, C, D, any_of(E, F, G)):
        out(Y1)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 9: A,B,C,D,any_of(E,F,G)", pyrung_to_ladder(logic, mapping))

# ── Case 10: nested any_of ──
with Program() as logic:
    with Rung(any_of(any_of(A, B), any_of(C, D))):
        out(Y1)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 10: any_of(any_of(A,B), any_of(C,D))", pyrung_to_ladder(logic, mapping))

# ── Case 11: negated OR branches ──
with Program() as logic:
    with Rung(A, any_of(~B, C, ~D)):
        out(Y1)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 11: A, any_of(~B, C, ~D)", pyrung_to_ladder(logic, mapping))

# ── Case 12: branch with pin rows ──
Done = Bool("Done")
Acc = Dint("Acc")
ResetCond = Bool("ResetCond")

with Program() as logic:
    with Rung(A, B):
        out(Y1)
        with branch(Mode):
            count_up(Done, Acc, preset=10).reset(ResetCond)

mapping = TagMap(
    {**base_map, Done: ct[1], Acc: ctd[1], ResetCond: x[8]},
    include_system=False,
)
_dump("Case 12: A,B out + branch(Mode) count_up.reset", pyrung_to_ladder(logic, mapping))

# ── Case 13: asymmetric OR widths ──
with Program() as logic:
    with Rung(any_of(A, all_of(B, C, D))):
        out(Y1)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 13: any_of(A, all_of(B,C,D))", pyrung_to_ladder(logic, mapping))

# ── Case 14: ORs sandwiching ANDs ──
with Program() as logic:
    with Rung(any_of(A, B), E, any_of(C, D), F):
        out(Y1)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 14: any_of(A,B), E, any_of(C,D), F", pyrung_to_ladder(logic, mapping))

# ── Case 15: OR + branch ──
with Program() as logic:
    with Rung(any_of(A, B)):
        out(Y1)
        with branch(Mode):
            out(Y2)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 15: any_of(A,B) + out + branch(Mode)", pyrung_to_ladder(logic, mapping))

# ── Case 16: 3-OR + branch ──
with Program() as logic:
    with Rung(any_of(A, B, C)):
        out(Y1)
        with branch(Mode):
            out(Y2)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 16: any_of(A,B,C) + out + branch(Mode)", pyrung_to_ladder(logic, mapping))

# ── Case 17: OR + two branches ──
with Program() as logic:
    with Rung(any_of(A, B)):
        out(Y1)
        with branch(Mode):
            out(Y2)
        with branch(Mode2):
            out(Y3)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 17: any_of(A,B) + out + branch(M1) + branch(M2)", pyrung_to_ladder(logic, mapping))

# ── Case 18: OR + branch + trailing out ──
with Program() as logic:
    with Rung(any_of(A, B)):
        out(Y1)
        with branch(Mode):
            out(Y2)
        out(Y3)

mapping = TagMap({**base_map}, include_system=False)
_dump("Case 18: any_of(A,B) + out + branch(Mode) + out", pyrung_to_ladder(logic, mapping))

# ── Case 19: OR + branch-first (no leading instruction) ──
with Program() as logic:
    with Rung(any_of(A, B)):
        with branch(Mode):
            out(Y1)
        with branch(Mode2):
            out(Y2)

mapping = TagMap({**base_map}, include_system=False)
_dump(
    "Case 19: any_of(A,B) + branch(M1) + branch(M2) [no leading out]",
    pyrung_to_ladder(logic, mapping),
)

print(f"\nCSVs written to: {OUT}")



"""Monitor/data-breakpoint ownership for DAP command handling.

Owns monitor CRUD/list events and dataBreakpoint* request handling.
Must preserve stop-reason semantics and event payload ordering.
"""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from pyrung.core.state import SystemState
from pyrung.dap.args import parse_args

HandlerResult = tuple[dict[str, Any], list[tuple[str, dict[str, Any] | None]]]


@dataclass
class MonitorMeta:
    id: int
    tag: str
    enabled: bool = True


@dataclass
class DataBreakpointMeta:
    data_id: str
    tag: str
    condition_source: str | None
    condition: Callable[[SystemState], bool] | None
    hit_condition: int | None
    hit_count: int = 0


@dataclass(frozen=True)
class _DataBreakpointInfoRequestArgs:
    variablesReference: Any = 0
    name: Any = None


@dataclass(frozen=True)
class _SetDataBreakpointsRequestArgs:
    breakpoints: Any = None


@dataclass(frozen=True)
class _SetDataBreakpointEntryArgs:
    dataId: Any = None
    condition: Any = None
    hitCondition: Any = None


@dataclass(frozen=True)
class _AddMonitorRequestArgs:
    tag: Any = None


@dataclass(frozen=True)
class _RemoveMonitorRequestArgs:
    id: Any = None


@dataclass(frozen=True)
class _FindLabelRequestArgs:
    label: Any = None
    all: Any = False


def on_data_breakpoint_info(adapter: Any, args: dict[str, Any]) -> HandlerResult:
    parsed = adapter._parse_request_args(_DataBreakpointInfoRequestArgs, args)
    variables_reference = int(parsed.variablesReference)
    name = parsed.name
    if variables_reference != adapter._monitor_scope_ref or not isinstance(name, str):
        return {
            "dataId": None,
            "description": "Data breakpoints are supported for PLC monitors",
        }, []

    with adapter._state_lock:
        monitor_tags = {meta.tag for meta in adapter._monitor_meta.values()}
    if name not in monitor_tags:
        return {"dataId": None, "description": f"No monitor registered for {name}"}, []

    data_id = adapter._data_id_for_tag(name)
    return {
        "dataId": data_id,
        "description": f"Break when {name} changes",
        "canPersist": True,
        "accessTypes": ["write"],
    }, []


def on_set_data_breakpoints(adapter: Any, args: dict[str, Any]) -> HandlerResult:
    parsed = adapter._parse_request_args(_SetDataBreakpointsRequestArgs, args)
    raw_breakpoints = parsed.breakpoints
    if not isinstance(raw_breakpoints, list):
        raw_breakpoints = []

    parsed_requests: list[tuple[int, dict[str, Any]]] = []
    responses: list[dict[str, Any] | None] = []
    for raw_bp in raw_breakpoints:
        if not isinstance(raw_bp, dict):
            responses.append({"verified": False, "message": "Breakpoint entry must be an object"})
            continue
        bp_args = parse_args(_SetDataBreakpointEntryArgs, raw_bp, error=adapter.DAPAdapterError)
        data_id = bp_args.dataId
        if not isinstance(data_id, str) or not data_id.strip():
            responses.append({"verified": False, "message": "dataId is required"})
            continue
        data_id = data_id.strip()
        condition_source = bp_args.condition
        if condition_source is not None and not isinstance(condition_source, str):
            responses.append({"verified": False, "message": "condition must be a string"})
            continue
        try:
            hit_condition = adapter._parse_hit_condition(bp_args.hitCondition)
        except adapter.DAPAdapterError as exc:
            responses.append({"verified": False, "message": str(exc)})
            continue
        parsed_requests.append(
            (
                len(responses),
                {
                    "dataId": data_id,
                    "condition": (
                        condition_source.strip() if isinstance(condition_source, str) else None
                    ),
                    "hitCondition": hit_condition,
                },
            )
        )
        responses.append(None)

    with adapter._state_lock:
        runner = adapter._require_runner_locked()
        requested_data_ids = {requested_bp["dataId"] for _, requested_bp in parsed_requests}
        stale = [
            data_id for data_id in adapter._data_bp_handles if data_id not in requested_data_ids
        ]
        for data_id in stale:
            handle = adapter._data_bp_handles.pop(data_id, None)
            if handle is not None:
                handle.remove()
            adapter._data_bp_meta.pop(data_id, None)

        for response_index, requested_bp in parsed_requests:
            data_id = requested_bp["dataId"]
            condition_source = requested_bp["condition"]
            existing_meta = adapter._data_bp_meta.get(data_id)
            unchanged = (
                existing_meta is not None
                and existing_meta.condition_source == condition_source
                and existing_meta.hit_condition == requested_bp["hitCondition"]
            )
            if unchanged:
                responses[response_index] = {"verified": True, "message": f"Watching {data_id}"}
                continue

            if data_id in adapter._data_bp_handles:
                handle = adapter._data_bp_handles.pop(data_id, None)
                if handle is not None:
                    handle.remove()
                adapter._data_bp_meta.pop(data_id, None)

            condition = None
            if condition_source:
                try:
                    condition = adapter._compile_condition(condition_source)
                except adapter.DAPAdapterError as exc:
                    responses[response_index] = {"verified": False, "message": str(exc)}
                    continue

            tag_name = adapter._tag_from_data_id(data_id)
            if tag_name is None:
                responses[response_index] = {
                    "verified": False,
                    "message": f"Unsupported dataId: {data_id}",
                }
                continue

            handle = runner.monitor(
                tag_name,
                adapter._build_data_breakpoint_callback(data_id=data_id),
            )
            adapter._data_bp_handles[data_id] = handle
            adapter._data_bp_meta[data_id] = DataBreakpointMeta(
                data_id=data_id,
                tag=tag_name,
                condition_source=condition_source,
                condition=condition,
                hit_condition=requested_bp["hitCondition"],
                hit_count=0,
            )
            responses[response_index] = {"verified": True, "message": f"Watching {tag_name}"}

    return {
        "breakpoints": [
            response
            if response is not None
            else {"verified": False, "message": "Invalid data breakpoint"}
            for response in responses
        ]
    }, []


def on_pyrung_add_monitor(adapter: Any, args: dict[str, Any]) -> HandlerResult:
    parsed = adapter._parse_request_args(_AddMonitorRequestArgs, args)
    tag = parsed.tag
    if not isinstance(tag, str) or not tag.strip():
        raise adapter.DAPAdapterError("pyrungAddMonitor.tag is required")
    tag_name = tag.strip()

    with adapter._state_lock:
        runner = adapter._require_runner_locked()
        monitor_id_ref: dict[str, int] = {"id": 0}
        handle = runner.monitor(
            tag_name,
            adapter._build_monitor_callback(tag_name=tag_name, monitor_id_ref=monitor_id_ref),
        )
        monitor_id_ref["id"] = handle.id
        adapter._monitor_handles[handle.id] = handle
        adapter._monitor_meta[handle.id] = MonitorMeta(id=handle.id, tag=tag_name, enabled=True)
        current = runner.current_state.tags.get(tag_name)
        adapter._monitor_values[handle.id] = adapter._format_value(current)
    return {"id": handle.id, "tag": tag_name, "enabled": True}, []


def on_pyrung_remove_monitor(adapter: Any, args: dict[str, Any]) -> HandlerResult:
    parsed = adapter._parse_request_args(_RemoveMonitorRequestArgs, args)
    raw_id = parsed.id
    if not isinstance(raw_id, int):
        raise adapter.DAPAdapterError("pyrungRemoveMonitor.id must be an integer")
    with adapter._state_lock:
        handle = adapter._monitor_handles.pop(raw_id, None)
        if handle is not None:
            handle.remove()
        removed = raw_id in adapter._monitor_meta
        adapter._monitor_meta.pop(raw_id, None)
        adapter._monitor_values.pop(raw_id, None)
    return {"id": raw_id, "removed": removed}, []


def on_pyrung_list_monitors(adapter: Any, _args: dict[str, Any]) -> HandlerResult:
    with adapter._state_lock:
        monitors = []
        for monitor_id in sorted(adapter._monitor_meta):
            meta = adapter._monitor_meta[monitor_id]
            monitors.append(
                {
                    "id": monitor_id,
                    "tag": meta.tag,
                    "enabled": bool(meta.enabled),
                    "value": adapter._monitor_values.get(monitor_id, "None"),
                }
            )
    return {"monitors": monitors}, []


def on_pyrung_find_label(adapter: Any, args: dict[str, Any]) -> HandlerResult:
    parsed = adapter._parse_request_args(_FindLabelRequestArgs, args)
    label = parsed.label
    if not isinstance(label, str) or not label.strip():
        raise adapter.DAPAdapterError("pyrungFindLabel.label is required")
    find_all = bool(parsed.all)

    with adapter._state_lock:
        runner = adapter._require_runner_locked()
        if find_all:
            matches = runner.history.find_all_labeled(label)
        else:
            latest = runner.history.find_labeled(label)
            matches = [] if latest is None else [latest]

    matches_payload = [
        {
            "scanId": match.scan_id,
            "timestamp": match.timestamp,
            "rtcIso": match.rtc_iso,
            "rtcOffsetSeconds": match.rtc_offset_seconds,
        }
        for match in matches
    ]
    return {"matches": matches_payload}, []


def monitor_variables(adapter: Any) -> list[dict[str, Any]]:
    variables: list[dict[str, Any]] = []
    for monitor_id in sorted(adapter._monitor_meta):
        meta = adapter._monitor_meta[monitor_id]
        tag_name = meta.tag
        variables.append(
            {
                "name": tag_name,
                "value": adapter._monitor_values.get(monitor_id, "None"),
                "type": "monitor",
                "evaluateName": tag_name,
                "variablesReference": 0,
            }
        )
    return variables


def build_monitor_callback(
    adapter: Any,
    *,
    tag_name: str,
    monitor_id_ref: dict[str, int],
) -> Callable[[Any, Any], None]:
    def _callback(current: Any, previous: Any) -> None:
        try:
            monitor_id = monitor_id_ref.get("id")
            if not isinstance(monitor_id, int) or monitor_id <= 0:
                return
            if monitor_id not in adapter._monitor_handles:
                return
            adapter._monitor_values[monitor_id] = adapter._format_value(current)
            runner = adapter._runner
            if runner is None:
                return
            state = runner.current_state
            adapter._enqueue_internal_event(
                "pyrungMonitor",
                {
                    "id": monitor_id,
                    "tag": tag_name,
                    "current": adapter._format_value(current),
                    "previous": adapter._format_value(previous),
                    "scanId": state.scan_id,
                    "timestamp": state.timestamp,
                },
            )
        except Exception:
            return

    return _callback


def build_data_breakpoint_callback(
    adapter: Any,
    *,
    data_id: str,
) -> Callable[[Any, Any], None]:
    def _callback(_current: Any, _previous: Any) -> None:
        try:
            meta = adapter._data_bp_meta.get(data_id)
            runner = adapter._runner
            if meta is None or runner is None:
                return

            condition = meta.condition
            if callable(condition) and not condition(runner.current_state):
                return

            hit_condition = meta.hit_condition
            hit_count = meta.hit_count + 1
            if hit_condition is None:
                meta.hit_count = hit_count
            else:
                if hit_count != int(hit_condition):
                    meta.hit_count = hit_count
                    return
                meta.hit_count = 0

            adapter._pending_predicate_pause = True
        except Exception:
            return

    return _callback


def data_id_for_tag(adapter: Any, tag_name: str) -> str:
    return f"tag:{tag_name}"


def tag_from_data_id(adapter: Any, data_id: str) -> str | None:
    prefix = "tag:"
    if not data_id.startswith(prefix):
        return None
    tag_name = data_id[len(prefix) :]
    return tag_name or None


def clear_debug_registrations_locked(adapter: Any) -> None:
    for handle in adapter._monitor_handles.values():
        try:
            handle.remove()
        except Exception:
            continue
    for handle in adapter._data_bp_handles.values():
        try:
            handle.remove()
        except Exception:
            continue
    adapter._monitor_handles.clear()
    adapter._monitor_meta.clear()
    adapter._monitor_values.clear()
    adapter._data_bp_handles.clear()
    adapter._data_bp_meta.clear()
    adapter._breakpoints.clear_source_breakpoints()
    adapter._pending_snapshot_labels_by_scan.clear()
    adapter._pending_predicate_pause = False
